#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from os import environ

from typing import NamedTuple

from healthcheckbot.common.error import ExpressionEvaluationError

__ALL__ = ('simple_env_evaluator',)

sel_expression_parser = re.compile(r'\$(?P<name>\w[\d\w]*)(?P<is_func>\((?P<arg>.*?)?\))?',
                                   re.IGNORECASE | re.MULTILINE)


class SelExpression(NamedTuple):
    name: str
    is_function: bool
    arguments: str


def simple_sel_parser(expression: str):
    def match_cb(m):
        return evaluate_expression(SelExpression(m.group('name'), m.group('is_func') is not None, m.group('arg')))

    return sel_expression_parser.sub(match_cb, expression)


def evaluate_expression(sel_expr: SelExpression) -> str:
    if sel_expr.name == 'env':
        return environ.get(sel_expr.arguments, '')
    else:
        raise ExpressionEvaluationError('Expression ${} is not supported'.format(sel_expr.name))


def simple_env_evaluator(original):
    if isinstance(original, str):
        return simple_sel_parser(original)
    return original


if __name__ == '__main__':
    simple_sel_parser('hello $env(MY_VAR1) and also $env(ANOTHER_VAR2)')



#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

def boolean(value):
    return isinstance(value, bool)


def integer(value):
    return isinstance(value, int)


def float_number(value):
    return isinstance(value, float)


def number(value):
    return isinstance(value, (int, float))


def string(value):
    return isinstance(value, str)


def dictionary(value):
    return isinstance(value, dict)


def dict_of_strings(value):
    if not dictionary(value):
        return False
    for v in value.values():
        if not isinstance(v, str):
            return False
    return True



#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

import sys
from typing import Callable, List

import os
import yaml

from healthcheckbot.common.core import ApplicationManager
from healthcheckbot.common.error import ConfigValidationError
from healthcheckbot.common.model import Module, WatcherModule
from healthcheckbot.common.utils import EvaluatingConfigWrapper

logger = logging.getLogger('Bootstrap')

__all__ = ['read_config', 'bootstrap', 'bootstrap_from_cli']


def read_config(config_file) -> dict:
    if isinstance(config_file, str):
        config_file = open(config_file, mode='r')
    try:
        return yaml.load(config_file)
    finally:
        if hasattr(config_file, 'close'):
            config_file.close()


def bootstrap(config: dict) -> ApplicationManager:
    application = bootstrap_environment(config)
    return bootstrap_from_cli(config, application)


def bootstrap_from_cli(config: dict, application: ApplicationManager) -> ApplicationManager:
    # Instantiate components
    instantiate_outputs(config, application)
    return application


def bootstrap_cli(config: dict) -> ApplicationManager:
    return bootstrap_environment(config, True)


def bootstrap_environment(config: dict, enable_cli=False) -> ApplicationManager:
    application = new_application()
    logger.info('Reading config file')
    wrapped_config = EvaluatingConfigWrapper(config)
    save_instance_config(wrapped_config, application)
    load_context_path(application)
    return application


def new_application(enable_cli=False):
    application = ApplicationManager()
    application.get_instance_settings().enable_cli = enable_cli
    return application


def save_instance_config(config: dict, application: ApplicationManager):
    if 'app' in config:
        app_config = config.get('app')
        settings = application.get_instance_settings()
        settings.id = app_config.get('id', 'healthcheckbot0')
        # Context Path
        for path in app_config.get('classpath', []):
            if isinstance(path, str):
                if not os.path.exists(path):
                    raise ConfigValidationError('app/classpath', 'path ' + path + ' doesn\'t exist')
                settings.context_path.append(path)
            else:
                raise ConfigValidationError('app/classpath', 'Context path must be the list of strings')


def load_context_path(application: ApplicationManager):
    context_path = application.get_instance_settings().context_path
    for p in context_path:
        sys.path.append(p)


def instantiate_modules_for_section(section_name: str, section_config: dict, application: ApplicationManager,
                                    cb: Callable[[ApplicationManager, Module, dict], None] = None,
                                    instance_name_prefix='') -> List[Module]:
    result = []
    for name, module_def in section_config.items():
        if not (isinstance(module_def, dict) and 'provider' in module_def):
            raise ConfigValidationError('/'.join((section_name, name)),
                                        'Should be dictionary containing mandatory key "provider"')
        instance = application.register_module_instance(instance_name_prefix + name, module_def.get('provider'))
        try:
            # Parse and set parameters
            for param_def in instance.PARAMS:
                if param_def.name in module_def:
                    val = module_def[param_def.name]
                    if param_def.parser is not None:
                        val = param_def.parser.parse(val, application, '/'.join((section_name, name, param_def.name)))
                    val = param_def.sanitize(val)
                    param_def.validate(val)
                    setattr(instance, param_def.name, val)
                elif param_def.is_required:
                    raise ConfigValidationError('/'.join((section_name, name, param_def.name)),
                                                'Parameter {} is required'.format(param_def.name))
            # Run validation of the module device
            instance.validate()
            if cb is not None:
                cb(application, instance, module_def)
            instance.on_configured()
            logger.info('Configured module {}({}))'.format(name, instance.__class__.__name__))
            result.append(instance)
        except Exception as e:
            raise ConfigValidationError('/'.join((section_name, name)), "Invalid module configuration: " + str(e), e)
    return result


def _register_watcher_module(application: ApplicationManager, watcher: WatcherModule, module_def: dict):
    for trigger_name in module_def.get('triggers', tuple()):
        trigger = application.get_trigger_by_name(trigger_name)
        if trigger is None:
            raise ConfigValidationError('triggers', "Trigger \'{}\' doesn't exist".format(trigger_name))
        trigger.register_watcher(watcher)
    assertions = instantiate_modules_for_section('watchers/{}/custom_assertions'.format(watcher.name),
                                                 module_def.get('custom_assertions', dict()), application,
                                                 instance_name_prefix=watcher.name + '__')
    watcher.custom_assertions = assertions


def instantiate_outputs(config: dict, application: ApplicationManager):
    section_config = config.get('outputs', {})
    instantiate_modules_for_section('outputs', section_config, application)
    section_config = config.get('triggers', {})
    instantiate_modules_for_section('triggers', section_config, application)
    section_config = config.get('watchers', {})
    instantiate_modules_for_section('watchers', section_config, application, _register_watcher_module)



#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json
import logging

import typing
from argparse import ArgumentParser
from typing import List

import yaml


class CliExtension(object):
    """
    Allows Module to extend CLI interface
    """
    COMMAND_NAME = None
    COMMAND_DESCRIPTION = None

    def __init__(self, parser: ArgumentParser, application):
        """
        :type application: healthcheckbot.common.core.ApplicationManager
        """
        super().__init__()
        self.parser = parser
        self.__application_manager = application

    def get_application_manager(self):
        """

        :rtype: healthcheckbot.common.core.ApplicationManager
        """
        return self.__application_manager

    @classmethod
    def setup_parser(cls, parser: ArgumentParser):
        pass

    def handle(self, args):
        raise NotImplementedError()


class Serializer(object):
    def serialize(self, obj):
        raise NotImplementedError("This method needs to be implemented in child class")

    def deserialize(self, data):
        raise NotImplementedError("This method needs to be implemented in child class")


class JsonSerializer(Serializer):
    def serialize(self, obj: object):
        json.dumps(obj)

    def deserialize(self, data: str):
        return json.loads(data)


class ConfigParser(object):
    def parse(self, config_section: dict, application, absolute_path: str = ''):
        """
        :param config_section:
        :param absolute_path: String containing absolute path from the config root to the given section.
        E.g. watchers/my_watcher/option1
        :type application: healthcheckbot.common.core.ApplicationManager
        """
        raise NotImplementedError()


class ParameterDef:
    def __init__(self, name: str, is_required=False, validators=None, parser: ConfigParser = None,
                 sanitize_fn: typing.Callable[[object], object] = None):
        super().__init__()
        self.name = name
        self.is_required = is_required
        self.validators = validators
        self.parser = parser
        self.sanitize_fn = sanitize_fn
        if self.validators is None:
            self.validators = []

    def sanitize(self, value):
        if self.sanitize_fn is not None:
            return self.sanitize_fn(value)
        return value

    def validate(self, value):
        for x in self.validators:
            if not x(value):
                raise ValueError('Parameter {} is invalid'.format(self.name))


class Module:
    PARAMS = []  # type: List[ParameterDef]
    MODULE_LEVEL_PARAMS = []  # type: List[ParameterDef]

    def __init__(self, application):
        """
        :type application: healthcheckbot.common.core.ApplicationManager
        """
        self.name = None
        self.__application = application
        self.logger = logging.getLogger(self.__class__.__name__)

    def get_application_manager(self):
        """
        :rtype application: healthcheckbot.common.core.ApplicationManager
        """
        return self.__application

    def on_initialized(self):
        pass

    def on_configured(self):
        pass

    def on_before_destroyed(self):
        pass

    def validate(self):
        pass


class LoopModuleMixin:
    """
    Modules inheriting this mixin will be registered in application loop. Thus step method will be invoked
    regularly as a part of main application routine.
    """

    def step(self):
        pass


class ValidationError(typing.NamedTuple):
    name: str
    description: str
    critical: bool


class WatcherResult:

    def __init__(self, state: dict = None, assertions_failed: List[ValidationError] = None) -> None:
        self.assertions_failed = assertions_failed
        self.state = state or {}

    def to_dict(self) -> typing.Dict:
        return {
            'failed_assertions': [x._asdict() for x in self.assertions_failed],
            'checks_passed': self.checks_passed,
            'state': self.state,
        }

    @property
    def checks_passed(self):
        return self.assertions_failed is None or len(self.assertions_failed) == 0


class ValidationReporter:

    def __init__(self, watcher, trigger) -> None:
        """
        :type watcher: healthcheckbot.common.core.WatcherModule
        :type trigger: healthcheckbot.common.core.TriggerModule
        """
        super().__init__()
        self.watcher = watcher
        self.trigger = trigger
        self.errors = []  # type: List[ValidationError]

    def error(self, name: str, description: str = None, critical=True):
        self.errors.append(ValidationError(name=name, description=description, critical=critical))


class WatcherModule(Module):

    def __init__(self, application):
        super().__init__(application)
        self.enable_assertions = True
        self.custom_assertions = []  # type: List[WatcherAssert]

    def obtain_state(self, trigger) -> object:
        """
        :type trigger: healthcheckbot.common.core.TriggerModule
        :return state object
        """
        pass

    def serialize_state(self, state: object) -> [dict, None]:
        if state is not None:
            return vars(state)
        else:
            return None

    def do_assertions(self, state: object, reporter: ValidationReporter):
        """
        :param reporter:
        :param state:
        :param trigger:
        """
        pass


class OutputModule(Module):

    def output(self, watcher_instance: WatcherModule, watcher_result: WatcherResult):
        pass


class TriggerModule(Module):

    def register_watcher(self, watcher: WatcherModule):
        pass


class WatcherAssert(Module):

    def do_assert(self, state: object, reporter: ValidationReporter, assertion_name: str):
        """
        :param assertion_name:
        :param reporter:
        :param state:
        :param trigger:
        """
        pass



#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

import traceback
from typing import Callable

import collections
import sys


class EvaluatingConfigWrapper(dict, collections.UserDict):

    def __init__(self, source: dict, evaluator: Callable[[str], str] = None):
        super().__init__(source)
        self.evaluator = evaluator

    def __getitem__(self, key):
        original = self.data.get(key)
        if isinstance(original, str) and self.evaluator is not None:
            return self.evaluator(original)
        return original



class CLI:
    verbose_mode = False

    @classmethod
    def print_data(cls, msg: str):
        print(msg)

    @classmethod
    def print_info(cls, msg: str):
        print(msg, file=sys.stderr)

    @classmethod
    def print_error(cls, exception):
        print(" -> ERROR: " + str(exception), file=sys.stderr)
        if cls.verbose_mode:
            print('--------------')
            traceback.print_exc(file=sys.stderr)

    @classmethod
    def print_debug(cls, string_to_print):
        if cls.verbose_mode:
            print('[DEBUG] ' + string_to_print, file=sys.stderr)



#    Healthcheck Bot
#    Copyright (C) 2018 Dmitry Berezovsky
#    
#    HealthcheckBot is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#    
#    HealthcheckBot is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import NamedTuple, Callable, List

import datetime

from healthcheckbot.common import validators
from healthcheckbot.common.model import TriggerModule, WatcherModule, LoopModuleMixin, ParameterDef


class SimpleTimerJob:

    def __init__(self, next_run: datetime.datetime = None, watcher: WatcherModule = None) -> None:
        self.next_run = next_run
        self.watcher = watcher


class SimpleTimer(TriggerModule, LoopModuleMixin):

    def __init__(self, application):
        super().__init__(application)
        self.interval = 300
        self.start_immediately = True
        self.__jobs = []  # type: List[SimpleTimerJob]

    def on_configured(self):
        pass

    def register_watcher(self, watcher: WatcherModule):
        self.__jobs.append(SimpleTimerJob(
            next_run=datetime.datetime.now() + datetime.timedelta(
                seconds=0 if self.start_immediately else self.interval),
            watcher=watcher
        ))

    def step(self):
        current_time = datetime.datetime.now()
        for job in self.__jobs:
            if job.next_run <= current_time:
                self.logger.info("Running watcher {}".format(job.watcher.name))
                self.get_application_manager().run_watcher(job.watcher, self)
                job.next_run = datetime.datetime.now() + datetime.timedelta(seconds=self.interval)

    PARAMS = (
        ParameterDef('interval', validators=(validators.integer,)),
        ParameterDef('start_immediately', validators=(validators.boolean,)),
    )
